
/*
 * tcl_blob.cc
 * Copyright (C) 1999 by John Heidemann
 * $Id: tcl_blob.cc,v 1.18 1999/08/07 07:06:02 johnh Exp $
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 */

#ifdef USE_TCL_BLOB

#include <stdlib.h>  // atoi
#include <tcl.h>
#include <iostream.h>

#include <tcl.h>
#include <tk.h>

#include "main.hh"
#include "tcl_blob.hh"
#include "process_list.hh"

#include <algorithm>


#include "gen.hh"


// instantiate me
blob *
blob::create_real_blob(process_view *pv)
{
	ENTRY_TRACE(__FILE__,__LINE__);
	return (blob*) new tcl_blob(pv);
}




// static stuff
Tcl_Interp *tcl_blob::interp_ = NULL;
char *tcl_blob::canvas_ = NULL;
tcl_blob::tcl_blobs_t tcl_blob::tcl_blobs_;
char tcl_blob::result_buf_[TCL_BUF_LEN];

static process_list pl;


// static
int
tcl_blob::lava_tick_proc(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
	ENTRY_TRACE(__FILE__,__LINE__);
	if (argc != 1)
		return TCL_ERROR;
	pl.scan();
	return TCL_OK;
}

// static
int
tcl_blob::lava_dump_proc(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
	ENTRY_TRACE(__FILE__,__LINE__);
	pl.dump();
	return TCL_OK;
}


// static
int
tcl_blob::lava_version_proc(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
	ENTRY_TRACE(__FILE__,__LINE__);
	strncpy(result_buf_, VERSION, TCL_BUF_LEN);
	interp->result = result_buf_;
	return TCL_OK;
}


// xxx: probably a better way to do this using bind2nd or something,
// but that requires too many brain cells.
struct blobs_find_id_p : public unary_function<tcl_blob *,bool> {
	blobs_find_id_p(int id) : id_(id) {};
	int id_;
	bool operator()(tcl_blob *p) { return p->id_ == id_; };
};

// static
int
tcl_blob::lava_id_to_info_proc(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
	ENTRY_TRACE(__FILE__,__LINE__);
	if (argc != 2)
		return TCL_ERROR;
	int id = atoi(argv[1]);
	// find it
	tcl_blobs_t::iterator result = find_if(tcl_blobs_.begin(), tcl_blobs_.end(), blobs_find_id_p(id));
	if (result == tcl_blobs_.end())
		return TCL_ERROR;
	// copy result
	(*result)->info_to_buf(result_buf_, TCL_BUF_LEN);
	interp->result = result_buf_;
	return TCL_OK;
}

// static
int
tcl_blob::lava_menu_proc(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
	ENTRY_TRACE(__FILE__,__LINE__);
	if (argc != 3)
		return TCL_ERROR;
	char *key = argv[1];
	char *value = argv[2];
	if (strcmp(key, "who") == 0) {
		filter_by_uids = (strcmp(value, "me") == 0);
	} else if (strcmp(key, "mem") == 0) {
		report_vm.tick_set(strcmp(value, "virtual") == 0);
	} else {
		return TCL_ERROR;
	};
	return TCL_OK;
}

// static
void
tcl_blob::init()
{
	ENTRY_TRACE(__FILE__,__LINE__);
	// const char* appname = "lavaps";
	interp_ = Tcl_CreateInterp();
	if (Tcl_Init(interp_) == TCL_ERROR)
		die(interp_->result);
	Tk_Window tk = NULL;
	if (Tk_Init(interp_) == TCL_OK)
		tk = Tk_MainWindow(interp_);
	if (tk == NULL)
		die("cannot start tk");

	/*
	 * add our new commands
	 */
	(void) Tcl_CreateCommand(interp_, "lava_tick", lava_tick_proc, NULL, NULL);
	(void) Tcl_CreateCommand(interp_, "lava_id_to_info", lava_id_to_info_proc, NULL, NULL);
	(void) Tcl_CreateCommand(interp_, "lava_menu", lava_menu_proc, NULL, NULL);
	(void) Tcl_CreateCommand(interp_, "lava_dump", lava_dump_proc, NULL, NULL);
	(void) Tcl_CreateCommand(interp_, "lava_version", lava_version_proc, NULL, NULL);
	// (void) Tcl_CreateCommand(interp_, "lava_move", lava_move_proc, NULL, NULL);
	// (void) Tcl_CreateCommand(interp_, "lava_grow", lava_grow_proc, NULL, NULL);
	// (void) Tcl_CreateCommand(interp_, "lava_print", lava_print_proc, NULL, NULL);
	// (void) Tcl_CreateCommand(interp_, "lava_reverse", lava_reverse_proc, NULL, NULL);

#if 0
	if (!b) {
		b = new tcl_blob(10, interp_, ".c");
		cout << "b created\n";
	};
#endif /* 0 */

	/*
	 * source some files
	 */
	if (TCL_OK != Tcl_Eval(interp_, (char *) lava_code)) {
		cerr << "error evaling bootstrap: " << interp_->result << endl;
		exit(1);
	};

	canvas_ = ".c";
}


// static
void
tcl_blob::mainloop()
{
	ENTRY_TRACE(__FILE__,__LINE__);
	Tk_MainLoop();
}


tcl_blob::tcl_blob(process_view *pv) :
	blob(pv), id_(0)
{
	ENTRY_TRACE(__FILE__,__LINE__);
	// xxx: assume blob initializer called first!
	tcl_blobs_.push_front(this);
	me_in_blobs_ = tcl_blobs_.begin();
	assert(me_in_blobs_ != tcl_blobs_.end());
	bd_color_ = "white";
	fill_color_ = "black";
	update_canvas();
}

tcl_blob::~tcl_blob()
{
	ENTRY_TRACE(__FILE__,__LINE__);
	char *bufp = result_buf_;

#if 0
	tcl_blobs_t::iterator result = find_if(tcl_blobs_.begin(), tcl_blobs_.end(), blobs_find_id_p(this->id_));
	assert(result != tcl_blobs_.end());
	tcl_blobs_.erase(result);
#else /* 0 */
	tcl_blobs_.erase(me_in_blobs_);
#endif /* 0 */

	if (id_) {
		sprintf(bufp, "%s delete %d\n", canvas_, id_);   bufp += strlen(bufp);
	};
	if (TCL_OK != Tcl_Eval(interp_, result_buf_)) {
		cerr << result_buf_ << endl << "tcl_blob::~tcl_blob: eval error\n";;
		exit (1);
	};
	id_ = 0;
}

void
tcl_blob::update_canvas()
{
	ENTRY_TRACE(__FILE__,__LINE__);
	char buf[LONG_TCL_BUF_LEN], *bufp = buf;
#define eobuf &buf[LONG_TCL_BUF_LEN]
	if (id_) {
		snprintf(bufp, eobuf-bufp, "%s delete %d\n", canvas_, id_);   bufp += strlen(bufp);
	};
	snprintf(bufp, eobuf-bufp, "set id [%s create polygon ", canvas_);   bufp += strlen(bufp);
	int i, x;
	for (i = 0, x = x_; i < num_; i++, x += x_step_) {
		snprintf(bufp, eobuf-bufp, "%d %d ", x, y_lows_[i]);   bufp += strlen(bufp);
	};
	for (; --i >= 0; ) {
		 x -= x_step_;
		snprintf(bufp, eobuf-bufp, "%d %d ", x, y_highs_[i]);   bufp += strlen(bufp);
	};
	// must always have 3 points
	if (num_ == 1) {
		// make one up
		x = x_ + x_step_ / 2 * (right_tending_ ? 1 : -1);
		int y = (y_highs_[0] - y_lows_[0]) / 2 + y_lows_[0];
		snprintf(bufp, eobuf-bufp, "%d %d ", x, y);   bufp += strlen(bufp);
	};
	snprintf(bufp, eobuf-bufp, "-fill %s -outline %s -smooth 1 -tag p]\n", fill_color_.c_str(), bd_color_.c_str());   bufp += strlen(bufp);

	blob *b = superior();
	int superior_id = b ? ((tcl_blob*)b)->id_ : 0;
	if (superior_id)
		snprintf(bufp, eobuf-bufp, "%s lower $id %d\n", canvas_, superior_id);   bufp += strlen(bufp);
	snprintf(bufp, eobuf-bufp, "return $id\n");   bufp += strlen(bufp);

	if (eobuf - bufp < 5) {
		int cont = 0;
		cerr << "tcl_blob::update_canvas: buffer overrun.\n";
		if (!cont)
			exit (1);
	};
	int e;
	if ((e = Tcl_Eval(interp_, buf)) != TCL_RETURN) {
		cerr << interp_->result << endl << "tcl_blob::update_canvas: eval error\nscript:\n" << buf << endl << endl;;
		exit (1);
	};
	id_ = atoi(interp_->result);
}

void
tcl_blob::op(blob_op op, int delta)
{
	ENTRY_TRACE(__FILE__,__LINE__);
	blob::op(op, delta);
	update_canvas();
}


void
tcl_blob::set_hsb(double h, double s, double b)
{
	ENTRY_TRACE(__FILE__,__LINE__);
	blob::set_hsb(h, s, b);
	const char *cs = hsb_to_rgbs(H_, S_, B_);
	fill_color_ = cs;
	free((void*)cs);
	cs = hsb_to_rgbs(H_, S_, 1.0 - B_);
	bd_color_ = cs;
	free((void*)cs);
	update_canvas();
}

#endif /* USE_TCL_BLOB */
